 /*
  *  TONKEY - Terminal Don't Key 
  *
  *  Create:   Mon Jul 29 00:50:09 1996  Susumu Sano
  *            Tue Dec  9 17:33:16 1997  Ver.0.4
  */

static char rcsid[] = "$Id: tonkey.c,v 1.10 1997/12/09 08:50:53 sano Exp $";
char	*version = "Ver.0.4";

#include <errno.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h> 
#include <sys/wait.h>

#include "config.h"

#ifndef CACHE_ENABLE_DEFAULT
#define CACHE_ENABLE_DEFAULT 0
#endif

#ifndef CTRL
#define CTRL(ch)   ((ch)&037)
#endif

#ifndef EWOULDBLOCK
#define EWOULDBLOCK EAGAIN
#endif

#ifndef _PATH_BSHELL
#define _PATH_BSHELL "/bin/sh"
#endif

char triggerkey = CTRL('\\');

char	*progname;

int	vflag;
int     cache  = CACHE_ENABLE_DEFAULT;
int	master, slave;
int	child;
#ifdef TIMEOUT /* NOT YET IMPLEMANT NOW*/
int     timeout = 60*60;  /* 1hour */
#endif
struct	termios tt;	/* saved terminal status */

void main PROTO((int, char**));
void doit PROTO((void));
void done PROTO((int));
void doshell PROTO((void));
void finish PROTO((int));

void saveoutput  PROTO((char *buf, int buflen));
static int comp PROTO((int pos, char *str));
static int otp_syntax PROTO((char *p));
static int findotpprompt PROTO((void));
static int readpass PROTO((char *prompt, char *phrase, int size));
static int trigger PROTO((void));
int findrequest PROTO((char *buf, int len));

void
usage()
{
	fprintf(stderr, "usage: %s [options...]\n", progname);
	fprintf(stderr, "options:\n");
	fprintf(stderr, "   -C : enable passphrase cache\n");
#ifdef TIMEOUT
	fprintf(stderr, "   -T sec : timeout of clear passphrase cache \n");
#endif
	fprintf(stderr, "%s %s\n", version, rcsid);
}

#ifdef  SIGWINCH
static void
windowchanged(int signo)
{
	struct winsize win;

	ioctl(0, TIOCGWINSZ, &win);
	ioctl(master, TIOCSWINSZ, &win);
}
#endif

void
main(argc, argv)
	int argc;
	char **argv;
{

	extern char *optarg;
	extern int optind, opterr;
	int  ch;

	progname = "tonkey";

	while ((ch = getopt(argc, argv, "vt:CT:")) != EOF)
		switch(ch) {
		case 'v':
			vflag = 1;
			break;
		case 't':
			triggerkey = atoi(optarg); 
			break;
		case 'C':
			cache = 1;
			break;
#ifdef TIMEOUT
		case 'T':	/* not yet implementation */
			timeout = atoi(optarg);
			break;
#endif
		case '?':
		default:
			usage();
			exit(1);
		}
	argc -= optind;
	argv += optind;

	printf("Start %s %s\n", progname, version);
	if (vflag)
		printf("\t%s\n", rcsid);

	fixtty();

	signal(SIGCHLD, finish);
	if ((child = fork()) == 0) {
		doshell();
		exit(1);
	}
	if (child == -1) {
		perror(progname);
		done(1);
	}
	close(slave);
	doit();
	done(0);
}

void
doit()
{
	int    n, nfds;
	int    in_cc, in_pos, out_cc, out_pos;
	fd_set r, w;
	char   in_buf[BUFSIZ], out_buf[BUFSIZ];
	struct timeval timeout;

	timeout.tv_sec = 5;
	timeout.tv_usec = 0;

	fcntl(0, F_SETFL, O_NONBLOCK);
	fcntl(1, F_SETFL, O_NONBLOCK);
	fcntl(master, F_SETFL, O_NONBLOCK);
#ifdef  SIGWINCH
 	signal(SIGWINCH, windowchanged);
#endif

	/* FD_ZERO(&r), FD_ZERO(&w); */
	in_cc = out_cc = 0;
	nfds = master + 1;

	while (in_cc >= 0 && out_cc >= 0) {
		FD_ZERO(&r),FD_ZERO(&w); /* I wish to move it to out
					    of the loop */
		if (in_cc == 0)
			FD_SET(0, &r);
		else if (in_cc > 0)
			FD_SET(master, &w);
		if (out_cc == 0)
			FD_SET(master, &r);
		else if (out_cc > 0)
			FD_SET(1, &w);

		if ((n = select(nfds, &r, &w, NULL, &timeout)) == -1 &&
		    errno != EINTR)
			break;

		/*printf("<%d %d %d> ", n, in_cc, out_cc); fflush(stdout); */

		if (FD_ISSET(0, &r)) {
			/* printf("inR");*/
			n = read(0, in_buf, sizeof(in_buf));
			if (n > 0) {
				if (findrequest(in_buf, n) == 0)
					in_cc = n, in_pos = 0;
			} else
				if (n == 0 || n == -1 && errno != EWOULDBLOCK) {
					in_cc = -1;
					/* printf("inR errno:%d\r\n", errno); /**/
				}
		}
		if (FD_ISSET(master, &r)) {
			/* printf("outR"); */
			n = read(master, out_buf, sizeof(out_buf));
			if (n > 0) {
				out_cc = n, out_pos = 0;
				saveoutput(out_buf, n);
			} else
				if (n == 0 || n == -1 && errno != EWOULDBLOCK) {
					out_cc = -1;
					/* printf("outR errno:%d\r\n", errno); */
				}
		}

		if (FD_ISSET(master, &w)) {
			/* printf("inW"); */
			n = write(master, &in_buf[in_pos], in_cc);
			if (n > 0)
				in_pos += n, in_cc -= n;
			if (n == 0 || n == -1 && errno != EWOULDBLOCK) {
				in_cc = -1;
					/* printf("inW errno:%d\r\n", errno); */
				}
		}
		
		if (out_cc > 0 && FD_ISSET(1, &w)) {
			/* printf("outW"); */
			n = write(1, &out_buf[out_pos], out_cc);
			if (n > 0)
				out_pos += n, out_cc -= n;
			if ( n == 0 || n == -1 && errno != EWOULDBLOCK) {
				out_cc = -1;
				/* printf("outW errno:%d\r\n", errno); */
			}
		}
	}
	fcntl(0, F_SETFL, 0);
	fcntl(1, F_SETFL, 0);
}

void
doshell()
{
	char *shell;

	if ((shell = getenv("SHELL")) == NULL)
		shell = _PATH_BSHELL;
	close(master);
	login_tty(slave);

#ifdef HAVE_SETENV
{
	/*
	 *  setenv tonkey ${mypid}
	 */
	char pid[16];
	sprintf(pid, "%d", getpid());
	setenv(progname, pid, 1);
	
	/* printf("start %s pid:%d, slave:%d\n", shell, getpid(), slave); */
}
#endif

	execl(shell, "sh", "-i", NULL);
	perror(shell);
	exit(1);
}

void
done(stat)
{

	fcntl(0, F_SETFL, 0);
	fcntl(1, F_SETFL, 0);
	restoretty();
	if (stat)
		printf("Abort %s %d\n", progname, stat);
	else
		printf("Finish %s\n", progname);
	exit(stat);
}

void
finish(signo)
{
	int die, pid;
	union wait status;

	die = 0;
	while ((pid = wait3((int *)&status, WNOHANG, 0)) > 0)
		if (pid == child)
			die = 1;
	if (die)
		done(0);
}

fixtty()
{
	struct termios rtt;
	struct winsize win;

	tcgetattr(0, &tt);
	ioctl(0, TIOCGWINSZ, &win);

	if (openpty(&master, &slave, NULL, &tt, &win) == -1) {
		if (errno == ENOENT)
			fprintf(stderr, "%s: no more pseudo terminal driver", progname);
		else
			perror("openpty"); /* unknown error */
		exit(1);
	}
	rtt = tt;
	cfmakeraw(&rtt);

	rtt.c_lflag &= ~ECHO;
	tcsetattr(0, TCSAFLUSH, &rtt);
}

restoretty()
{
	tcsetattr(0, TCSAFLUSH, &tt);
}


/*
 * From md.h
 */
#define MD2 0
#define MD4 1
#define MD5 2

int MD;

#define min(a,b) ((a)<(b)?(a):(b))

extern char *btoe PROTO((char *));

int   savebuflen = 512;
char  *savebuf;

static int  savep;

static void
dump() {
	int i, ch, p = savep;
	printf("\r\nsavep:%d ", savep);
	for (i = 0; i < 200; i++) {
		if (--p < 0)
			p = savebuflen-1;
		ch = savebuf[p];
		if (ch < ' ' || ch > '~')
			printf("\\%0o", ch);
		else
			printf("%c", ch);
	}
	printf("\r\n");
}

void
saveoutput(buf, buflen)
	char *buf;
	int buflen;
{
	if (savebuf == NULL) {
		if ((savebuf = (char*)malloc(savebuflen)) == NULL) {
			perror(progname);
			done(1);
		}
		savep = 0;
	}
	if (buflen >= savebuflen) {
		memcpy(savebuf, &buf[buflen-savebuflen], savebuflen);
		savep = 0;
	} else {
		int len = min(buflen, savebuflen - savep);
		memcpy(&savebuf[savep], buf, len);
		savep += len;
		if (savep >= savebuflen) {
			assert(savep == savebuflen);
			savep = 0;
		}
		if (len < buflen) {
			memcpy(savebuf, &buf[len], buflen - len);
			savep = buflen - len;
		}
	}
}

struct {
	char *key;
	int  func;
} keys[] = {
        { "\ns/key",   MD4 },
	{ "\notp-md2", MD2 },
	{ "\notp-md4", MD4 },
	{ "\notp-md5", MD5 },
        { "\nS/Key MD2",   MD2 },
        { "\nS/Key MD4",   MD4 },
        { "\nS/Key MD5",   MD5 },
        { "\nS/Key (MD2)",   MD2 },
        { "\nS/Key (MD4)",   MD4 },
        { "\nS/Key (MD5)",   MD5 },
        { "\n331 s/key",   MD4 },
	{ "\n331 otp-md2", MD2 },
	{ "\n331 otp-md4", MD4 },
	{ "\n331 otp-md5", MD5 },
	{ "\n331 OTP response otp-md2", MD2 },
	{ "\n331 OTP response otp-md4", MD4 },
	{ "\n331 OTP response otp-md5", MD5 },
	{ "\totp-md2", MD2 },
	{ "\totp-md4", MD4 },
	{ "\totp-md5", MD5 },
	{ NULL, 0},
};

static int func;
static int seq;
static char *seed;

#define next(pos, inc) if (((pos)+=(inc)) >= savebuflen) pos = 0

static int 
comp(pos, str)
	int pos;
	char *str;
{
	int c;
	while (*str) {
		if ((c = savebuf[pos] - *str++) != 0)
			return c;
		next(pos, 1);
	}
	return 0;
}

static int
otp_syntax(p)
	char *p;
{
	if (*p == 0)
		return 0;
	while (isspace(*p))
		p++;
	if (!isdigit(*p))
		return 0;
	seq = 0;
	do
		seq = 10 * seq + (*p++ - '0');
	while (isdigit(*p));
	if (*p == 0 || !isspace(*p))
		return 0;
	while (isspace(*p))
		p++;
	if (*p == 0)
		return 0;
	seed = p;
	while (*p && !isspace(*p))
		p++;
	*p = 0;
	return 1;
}
 
static int
findotpprompt()
{
	int pos, k, i;
	static char line[128];

	pos = savep;
	savebuf[pos] = 0;
	for (;;) {
		if (--pos < 0)
			pos = savebuflen - 1;
		if (pos == savep)
			return (0);
		for (k = 0; keys[k].key; k++)
			if (comp(pos, keys[k].key) == 0) {
				int savepos = pos;
				next(pos, 1);
				for (i = 0; i < sizeof(line)-1 && pos != savep;) {
					int ch = savebuf[pos];
					if (ch == '\r' || ch == '\n')
						break;
					line[i++] = savebuf[pos];
					next(pos, 1);
				}
				line[i] = 0;
				func = keys[k].func;
				if (otp_syntax(line + strlen(keys[k].key)))
					return 1;
				pos = savepos;
			}
	}
}


void
msg(fd, str, canwait)
	int fd;
	char *str;
	int canwait;
{
	if (canwait) {
		fcntl(fd, F_SETFL, 0);
		write(fd, str, strlen(str));
		fcntl(fd, F_SETFL, O_NONBLOCK);
	} else
		write(fd, str, strlen(str));
}

/*
 * Read pass phrase from terminal. We assume mode of the terminal be
 * no-echo and raw mode.
 */
static int
readpass(prompt, phrase, size)
	char *prompt;
	char *phrase;
	int size;
{
	char ch;
	int pos, err;
	int echo = vflag;

 	fcntl(0, F_SETFL, 0);
 	fcntl(1, F_SETFL, 0);

	msg(1, "\r\n", 0);
	msg(1, prompt, 0);

	pos = err = 0;
	for (;;) {
		read(0, &ch, 1);
		ch &= 0177;
		switch (ch){
		case CTRL('c'):
		case CTRL('z'):
			err = -1;
			msg(1, "*BREAK*\r\n", 0);
			goto e;
		case '\b':
		case 0177:
			if (pos > 0) {
				pos--;
				if (echo) 
					msg(1, "\b \b", 0);
			}
			break;
		case CTRL('u'):
			while (pos > 0) {
				pos--;
				if (echo)
					msg(1, "\b \b", 0);
			}
			break;
		case '\r':
		case '\n':
			phrase[pos] = 0;
			msg(1, "\r\n", 0);
			goto e;
		default:
			if (pos < size-1) {
				phrase[pos++] = ch;
				if (echo)
					msg(1, "*", 0);
			}
		}
	}
 e:
 	fcntl(0, F_SETFL, O_NONBLOCK);
 	fcntl(1, F_SETFL, O_NONBLOCK);
	return err;
}

static int
trigger()
{
	/* 
	 * TODO: encrypt & save
	 */
	static char  save_passphrase[PASS_PHRASE_MAX_LEN]; 

	/* printf("\r\n!!\r\n"); dump(); /**/
	if (findotpprompt()) {
		char key[8];
		char  passphrase[PASS_PHRASE_MAX_LEN];
		char  *s;
		int i;

		if (readpass("Enter passphrase : ", passphrase, sizeof(passphrase)) == -1)
			return 0;
		if (cache) {
			if (*passphrase == 0)
				strcpy(passphrase, save_passphrase);
			strcpy(save_passphrase, passphrase);
		}
		if (*passphrase == 0) {
			msg(1, "*No Passphrase*\r\n", 1);
			return 0;
		}

		MD = func;

		if (keycrunch(key, seed, passphrase) == RET_ERROR) {
			memset(passphrase, 0, sizeof(passphrase));
			return 0;
		}
		memset(passphrase, 0, sizeof(passphrase));

		for (i = 0; i < seq; i++)
			secure_hash(key);

		s = btoe(key);
		
		msg(master, s, 1); msg(master, "\r", 1);
		msg(1, s, 1); msg(1, "\r\n", 1);

		return 1;
	}
	return 0;
}

int
findrequest(buf, len)
	char *buf;
	int len;
{
	if (len == 1 && *buf == triggerkey)
		return trigger();
	return 0;
}

/*
 * Local Variables:
 *  c-indent-level:               8
 *  c-continued-statement-offset: 8
 *  c-continued-brace-offset:     0
 *  c-brace-offset:              -8
 *  c-brace-imaginary-offset:     0
 *  c-argdecl-indent:             8
 *  c-label-offset:              -8
 *  c++-hanging-braces:           1
 *  c++-access-specifier-offset: -8
 *  c++-empty-arglist-indent:     8
 *  c++-friend-offset:            0
 * End:
 */
